--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2007 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <stefan.fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH THE CESYS EFM01
-- USB CARD (ARTICLE-NR.: C1050-4107) ONLY! YOU ARE NOT ALLOWED TO MODIFY AND
-- DISTRIBUTE OR USE IT WITH ANY OTHER HARDWARE, SOFTWARE OR ANY OTHER KIND OF
-- ASIC OR PROGRAMMABLE LOGIC DESIGN WITHOUT THE EXPLICIT PERMISSION OF THE
-- COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

-- pragma translate_off
library UNISIM;
use UNISIM.VCOMPONENTS.all;
-- pragma translate_on


entity sfifo_hd_a1Kx18b0K5x36 is
  generic
  (
    nr_of_dbgports : positive := 1
  );
  port
  (
    rst : in std_logic;
    clk : in std_logic;
    
    dir_a2b_b2a_n_i : in std_logic;

    a_wr_i : in std_logic;
    a_wr_data_i : in std_logic_vector(17 downto 0);
    a_rd_i : in std_logic;
    a_rd_data_o : out std_logic_vector(17 downto 0);

    b_wr_i : in std_logic;
    b_wr_data_i : in std_logic_vector(35 downto 0);
    b_rd_i : in std_logic;
    b_rd_data_o : out std_logic_vector(35 downto 0);

    full_o : out std_logic;
    empty_o : out std_logic;
    count_o : out std_logic_vector(8 downto 0);

    debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
  );
end sfifo_hd_a1Kx18b0K5x36;


architecture RTL of sfifo_hd_a1Kx18b0K5x36 is

  signal dir_a2b_b2a_n : std_logic := '0';

  signal a_wr : std_logic := '0';
  signal a_wr_data : std_logic_vector(17 downto 0) := (others => '0');
  signal a_rd : std_logic := '0';
  signal a_rd_data : std_logic_vector(17 downto 0) := (others => '0');

  signal b_wr : std_logic := '0';
  signal b_wr_data : std_logic_vector(35 downto 0) := (others => '0');
  signal b_rd : std_logic := '0';
  signal b_rd_data : std_logic_vector(35 downto 0) := (others => '0');

  signal full : std_logic := '1';
  signal empty : std_logic := '1';
  signal count : std_logic_vector(8 downto 0) := (others => '0');

  component RAMB16_S18_S36
  -- pragma translate_off
    generic (
  -- "Read during Write" attribute for functional simulation
   WRITE_MODE_A : string := "READ_FIRST" ; -- WRITE_FIRST(default)/ READ_FIRST/ NO_CHANGE
   WRITE_MODE_B : string := "WRITE_FIRST" ; -- WRITE_FIRST(default)/ READ_FIRST/ NO_CHANGE
  -- RAM initialization ("0" by default) for functional simulation: see example
   );
  -- pragma translate_on
    port (
  	DIA     : in std_logic_vector (15 downto 0);
  	DIPA    : in std_logic_vector (1 downto 0);
  	ADDRA   : in std_logic_vector (9 downto 0);
  	ENA     : in std_logic;
  	WEA     : in std_logic;
  	SSRA    : in std_logic;
  	CLKA    : in std_logic;
  	DOA     : out std_logic_vector (15 downto 0);
  	DOPA    : out std_logic_vector (1 downto 0);	
  --
  	DIB     : in std_logic_vector (31 downto 0);
  	DIPB    : in std_logic_vector (3 downto 0);
  	ADDRB   : in std_logic_vector (8 downto 0);
  	ENB     : in std_logic;
  	WEB     : in std_logic;
  	SSRB    : in std_logic;
  	CLKB    : in std_logic;
  	DOB     : out std_logic_vector (31 downto 0);
  	DOPB    : out std_logic_vector (3 downto 0)	
  	); 
  end component;

  attribute WRITE_MODE_A : string;
  attribute WRITE_MODE_A of U_RAMB16_S18_S36: label is "READ_FIRST";
  attribute WRITE_MODE_B : string;
  attribute WRITE_MODE_B of U_RAMB16_S18_S36: label is "WRITE_FIRST"; 

	signal DIA     : std_logic_vector (15 downto 0) := (others => '0');
	signal DIPA    : std_logic_vector (17 downto 16) := (others => '0');
	signal ADDRA   : std_logic_vector (9 downto 0) := (others => '0');
	signal ENA     : std_logic := '0';
	signal WEA     : std_logic := '0';
	signal SSRA    : std_logic := '0';
	signal CLKA    : std_logic := '0';
	signal DOA     : std_logic_vector (15 downto 0) := (others => '0');
	signal DOPA    : std_logic_vector (17 downto 16) := (others => '0');
--
	signal DIB     : std_logic_vector (31 downto 0) := (others => '0');
	signal DIPB    : std_logic_vector (35 downto 32) := (others => '0');
	signal ADDRB   : std_logic_vector (8 downto 0) := (others => '0');
	signal ENB     : std_logic := '0';
	signal WEB     : std_logic := '0';
	signal SSRB    : std_logic := '0';
	signal CLKB    : std_logic := '0';
	signal DOB     : std_logic_vector (31 downto 0) := (others => '0');
	signal DOPB    : std_logic_vector (35 downto 32) := (others => '0');

  signal a_wr_en : std_logic := '0';
  signal a_wr_en_lsbs : std_logic := '0';
  signal a_rd_en : std_logic := '0';
  signal a_rd_en_lsbs : std_logic := '0';
  signal a_pointer : unsigned(9 downto 1) := (others => '0');
  signal a_pointer_lsbs : unsigned(0 downto 0) := (others => '0');

  signal b_wr_en : std_logic := '0';
  signal b_rd_en : std_logic := '0';
  signal b_pointer : unsigned(8 downto 0) := (others => '0');

begin

  dir_a2b_b2a_n <= dir_a2b_b2a_n_i;

  a_wr <= a_wr_i;
  a_wr_data <= a_wr_data_i;
  a_rd <= a_rd_i;
  a_rd_data_o <= a_rd_data;

  b_wr <= b_wr_i;
  b_wr_data <= b_wr_data_i;
  b_rd <= b_rd_i;
  b_rd_data_o <= b_rd_data;

  full_o <= full;
  empty_o <= empty;
  count_o <= count;
  
  debug <=
  (
    others => 'Z'
  );

  U_RAMB16_S18_S36: RAMB16_S18_S36
    port map (
    DIA     => DIA, -- insert 16 bits data in bus (<15 downto 0>)
  	DIPA    => DIPA, -- insert 2 bits parity data in bus (or <17 downto 16>)
  	ADDRA   => ADDRA, -- insert 10 bits address bus
  	ENA     => ENA, -- insert enable signal
  	WEA     => WEA, -- insert write enable signal
  	SSRA    => SSRA, -- insert set/reset signal
  	CLKA    => CLKA, -- insert clock signal
  	DOA     => DOA, -- insert 16 bits data out bus (<15 downto 0>)
  	DOPA    => DOPA, -- insert 2 bits parity data out bus (or <17 downto 16>)
  --
  	DIB     => DIB, -- insert 32 bits data in bus (<31 downto 0>)
  	DIPB    => DIPB, -- insert 4 bits parity data in bus (or <35 downto 32>)
  	ADDRB   => ADDRB, -- insert 9 bits address bus
  	ENB     => ENB, -- insert enable signal
  	WEB     => WEB, -- insert write enable signal
  	SSRB    => SSRB, -- insert set/reset signal
  	CLKB    => CLKB, -- insert clock signal
  	DOB     => DOB, -- insert 32 bits data out bus (<31 downto 0>)
  	DOPB    => DOPB  -- insert 4 bits parity data out bus (or <35 downto 32>)
  	);

	DIA     <= a_wr_data(15 downto 0);
	DIPA    <= a_wr_data(17 downto 16);
	ADDRA   <= std_logic_vector(a_pointer & a_pointer_lsbs);
	ENA     <= a_wr_en_lsbs or a_rd_en_lsbs;
	WEA     <= a_wr_en_lsbs;
	SSRA    <= rst;
	CLKA    <= clk;
	a_rd_data(15 downto 0) <= DOA;
	a_rd_data(17 downto 16) <= DOPA;
--
	DIB     <= b_wr_data(31 downto 0);
	DIPB    <= b_wr_data(35 downto 32);
	ADDRB   <= std_logic_vector(b_pointer);
	ENB     <= b_wr_en or b_rd_en;
	WEB     <= b_wr_en;
	SSRB    <= rst;
	CLKB    <= clk;
	b_rd_data(31 downto 0) <= DOB;
	b_rd_data(35 downto 32) <= DOPB;

  a_wr_en <= a_wr_en_lsbs and a_pointer_lsbs(0);
  a_rd_en <= a_rd_en_lsbs and a_pointer_lsbs(0);

  process(rst, dir_a2b_b2a_n, clk, a_wr, full, b_rd, empty, a_rd, b_wr)
    variable wr_rd : std_logic_vector(1 downto 0) := (others => '0');
  begin

    if rst = '1' then
      a_pointer_lsbs <= (others => '0');
      a_pointer <= (others => '0');
      b_pointer <= (others => '0');
      full <= '1';
      empty <= '1';
      count <= (others => '0');
    elsif clk'event and clk = '1' then
      if (a_wr_en_lsbs or a_rd_en_lsbs) = '1' then
        a_pointer_lsbs <= a_pointer_lsbs + 1;
      end if;
      if (a_wr_en or a_rd_en) = '1' then
        a_pointer <= a_pointer + 1;
      end if;
      if (b_wr_en or b_rd_en) = '1' then
        b_pointer <= b_pointer + 1;
      end if;
      if (unsigned(count) = 0) or (unsigned(count) = 1 and (a_rd_en or b_rd_en) = '1' and (a_wr_en or b_wr_en) = '0') then
        empty <= '1';
      else
        empty <= '0';
      end if;
      if (unsigned(count) = 511) or (unsigned(count) = 510 and (a_wr_en or b_wr_en) = '1' and (a_rd_en or b_rd_en) = '0') then
        full <= '1';
      else
        full <= '0';
      end if;
      wr_rd := (a_wr_en or b_wr_en) & (a_rd_en or b_rd_en);
      case wr_rd is
        when b"01" =>
          count <= std_logic_vector(unsigned(count) - 1);
        when b"10" =>
          count <= std_logic_vector(unsigned(count) + 1);
        when others => null;
      end case;
    end if;
    
    a_wr_en_lsbs <= '0';
    a_rd_en_lsbs <= '0';
    b_wr_en <= '0';
    b_rd_en <= '0';
    if dir_a2b_b2a_n = '1' then
      a_wr_en_lsbs <= a_wr and (not full);
      b_rd_en <= b_rd and (not empty);
    else
      a_rd_en_lsbs <= a_rd and (not empty);
      b_wr_en <= b_wr and (not full);
    end if;

  end process;

end RTL;