--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2007 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <stefan.fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH THE CESYS EFM01
-- USB CARD (ARTICLE-NR.: C1050-4107) ONLY! YOU ARE NOT ALLOWED TO MODIFY AND
-- DISTRIBUTE OR USE IT WITH ANY OTHER HARDWARE, SOFTWARE OR ANY OTHER KIND OF
-- ASIC OR PROGRAMMABLE LOGIC DESIGN WITHOUT THE EXPLICIT PERMISSION OF THE
-- COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.wishbone.all;


entity wb_intercon is
  generic
  (
    nr_of_dbgports : positive := 1
  );
  port
  (
    SYSCON_I : in rec_syscon_port;

    SYSCON_MA_O : out arr_syscon_port(rng_masters);
    MASTER_I : in arr_master_port(rng_masters);
    SLAVE_O : out arr_slave_port(rng_masters);

    SYSCON_SL_O : out arr_syscon_port(rng_slaves);
    MASTER_O : out arr_master_port(rng_slaves);
    SLAVE_I : in arr_slave_port(rng_slaves);

    debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
  );
end wb_intercon;


architecture RTL of wb_intercon is

  signal intercon : rec_intercon_signal := intercon_default;

  signal wishbone : rec_wishbone_signal := wishbone_default;

begin

  -- connect ports to internal signals
  ------------------------------------

  intercon.syscon <= SYSCON_I;

  SYSCON_MA_O <= intercon.masters.syscon;
  intercon.masters.master <= MASTER_I;
  SLAVE_O <= intercon.masters.slave;

  SYSCON_SL_O <= intercon.slaves.syscon;
  MASTER_O <= intercon.slaves.master;
  intercon.slaves.slave <= SLAVE_I;
  
  debug <=
  (
    others => 'Z'
  );

  -- architecture implementation
  ------------------------------

  -- distribute the same syscon signal (rst, clk) to all wishbone devices
  wishbone.syscon <= intercon.syscon;
  intercon.masters.syscon <= (others => wishbone.syscon);
  intercon.slaves.syscon <= (others => wishbone.syscon);
  
  -- the fx-2 usb controller is the one and only master at
  -- the on-chip-bus, no arbitration is needed
  -- connect this master to all slave devices
  wishbone.master <= intercon.masters.master(ma_fx2_id);
  intercon.masters.slave <= (others => wishbone.slave);

  -- building shared bus topology
  process(wishbone, intercon)
  begin
    -- slave mux
    wishbone.slave <= slave_default;
    intercon.slaves.master <= (others => wishbone.master);
    -- partial address decoding, using some msbs of address to
    -- select the appropriate slave device
    -- the for-loop represents the slave mux
    for i in rng_slaves loop
      -- cyc is the one and only bit in the master signal, which has to be muxed,
      -- you could mux the whole master signal too, but this would waste FPGA resources
      intercon.slaves.master(i).cyc <= '0'; -- default: slave is deselected
      if i = TO_INTEGER(unsigned(wishbone.master.adr(rng_slave_select))) then
        intercon.slaves.master(i).cyc <= '1';
        wishbone.slave <= intercon.slaves.slave(i);
      end if;
    end loop;
  end process;

end RTL;